@@ -2,8 +2,8 @@
  * libdpkg - Debian packaging suite library routines
  * parse.c - database file parsing, main package/field loop
  *
- * Copyright © 1995 Ian Jackson <ian@chiark.greenend.org.uk>
- * Copyright © 2006,2008-2014 Guillem Jover <guillem@debian.org>
+ * Copyright © 1995 Ian Jackson <ijackson@chiark.greenend.org.uk>
+ * Copyright © 2006, 2008-2015 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -28,9 +28,8 @@
 #include <sys/mman.h>
 #endif
 
-#include <assert.h>
+#include <errno.h>
 #include <fcntl.h>
-#include <ctype.h>
 #include <string.h>
 #include <unistd.h>
 #include <stdarg.h>
@@ -39,6 +38,7 @@
 
 #include <dpkg/macros.h>
 #include <dpkg/i18n.h>
+#include <dpkg/c-ctype.h>
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
 #include <dpkg/string.h>
@@ -52,50 +52,48 @@
  */
 const struct fieldinfo fieldinfos[]= {
   /* Note: Capitalization of field name strings is important. */
-  { "Package",          f_name,            w_name                                     },
-  { "Essential",        f_boolean,         w_booleandefno,   PKGIFPOFF(essential)     },
-  { "Status",           f_status,          w_status                                   },
-  { "Priority",         f_priority,        w_priority                                 },
-  { "Section",          f_section,         w_section                                  },
-  { "Installed-Size",   f_charfield,       w_charfield,      PKGIFPOFF(installedsize) },
-  { "Origin",           f_charfield,       w_charfield,      PKGIFPOFF(origin)        },
-  { "Maintainer",       f_charfield,       w_charfield,      PKGIFPOFF(maintainer)    },
-  { "Bugs",             f_charfield,       w_charfield,      PKGIFPOFF(bugs)          },
-  { "Architecture",     f_architecture,    w_architecture                             },
-  { "Multi-Arch",       f_multiarch,       w_multiarch,      PKGIFPOFF(multiarch)     },
-  { "Source",           f_charfield,       w_charfield,      PKGIFPOFF(source)        },
-  { "Version",          f_version,         w_version,        PKGIFPOFF(version)       },
-  { "Revision",         f_revision,        w_null                                     },
-  { "Config-Version",   f_configversion,   w_configversion                            },
-  { "Replaces",         f_dependency,      w_dependency,     dep_replaces             },
-  { "Provides",         f_dependency,      w_dependency,     dep_provides             },
-  { "Depends",          f_dependency,      w_dependency,     dep_depends              },
-  { "Pre-Depends",      f_dependency,      w_dependency,     dep_predepends           },
-  { "Recommends",       f_dependency,      w_dependency,     dep_recommends           },
-  { "Suggests",         f_dependency,      w_dependency,     dep_suggests             },
-  { "Breaks",           f_dependency,      w_dependency,     dep_breaks               },
-  { "Conflicts",        f_dependency,      w_dependency,     dep_conflicts            },
-  { "Enhances",         f_dependency,      w_dependency,     dep_enhances             },
-  { "Conffiles",        f_conffiles,       w_conffiles                                },
-  { "Filename",         f_filecharf,       w_filecharf,      FILEFOFF(name)           },
-  { "Size",             f_filecharf,       w_filecharf,      FILEFOFF(size)           },
-  { "MD5sum",           f_filecharf,       w_filecharf,      FILEFOFF(md5sum)         },
-  { "MSDOS-Filename",   f_filecharf,       w_filecharf,      FILEFOFF(msdosname)      },
-  { "Description",      f_charfield,       w_charfield,      PKGIFPOFF(description)   },
-  { "Triggers-Pending", f_trigpend,        w_trigpend                                 },
-  { "Triggers-Awaited", f_trigaw,          w_trigaw                                   },
-  /* Note that aliases are added to the nicknames table. */
-  {  NULL                                                                             }
-};
+  { FIELD("Package"),          f_name,            w_name                                     },
+  { FIELD("Essential"),        f_boolean,         w_booleandefno,   PKGIFPOFF(essential)     },
+  { FIELD("Protected"),        f_boolean,         w_booleandefno,   PKGIFPOFF(is_protected)  },
+  { FIELD("Status"),           f_status,          w_status                                   },
+  { FIELD("Priority"),         f_priority,        w_priority                                 },
+  { FIELD("Section"),          f_section,         w_section                                  },
+  { FIELD("Installed-Size"),   f_charfield,       w_charfield,      PKGIFPOFF(installedsize) },
+  { FIELD("Origin"),           f_charfield,       w_charfield,      PKGIFPOFF(origin)        },
+  { FIELD("Maintainer"),       f_charfield,       w_charfield,      PKGIFPOFF(maintainer)    },
+  { FIELD("Bugs"),             f_charfield,       w_charfield,      PKGIFPOFF(bugs)          },
+  { FIELD("Architecture"),     f_architecture,    w_architecture                             },
+  { FIELD("Multi-Arch"),       f_multiarch,       w_multiarch,      PKGIFPOFF(multiarch)     },
+  { FIELD("Source"),           f_charfield,       w_charfield,      PKGIFPOFF(source)        },
+  { FIELD("Version"),          f_version,         w_version,        PKGIFPOFF(version)       },
+  { FIELD("Config-Version"),   f_configversion,   w_configversion                            },
+  { FIELD("Replaces"),         f_dependency,      w_dependency,     dep_replaces             },
+  { FIELD("Provides"),         f_dependency,      w_dependency,     dep_provides             },
+  { FIELD("Depends"),          f_dependency,      w_dependency,     dep_depends              },
+  { FIELD("Pre-Depends"),      f_dependency,      w_dependency,     dep_predepends           },
+  { FIELD("Recommends"),       f_dependency,      w_dependency,     dep_recommends           },
+  { FIELD("Suggests"),         f_dependency,      w_dependency,     dep_suggests             },
+  { FIELD("Breaks"),           f_dependency,      w_dependency,     dep_breaks               },
+  { FIELD("Conflicts"),        f_dependency,      w_dependency,     dep_conflicts            },
+  { FIELD("Enhances"),         f_dependency,      w_dependency,     dep_enhances             },
+  { FIELD("Conffiles"),        f_conffiles,       w_conffiles                                },
+  { FIELD("Filename"),         f_archives,        w_archives,       ARCHIVEFOFF(name)        },
+  { FIELD("Size"),             f_archives,        w_archives,       ARCHIVEFOFF(size)        },
+  { FIELD("MD5sum"),           f_archives,        w_archives,       ARCHIVEFOFF(md5sum)      },
+  { FIELD("MSDOS-Filename"),   f_archives,        w_archives,       ARCHIVEFOFF(msdosname)   },
+  { FIELD("Description"),      f_charfield,       w_charfield,      PKGIFPOFF(description)   },
+  { FIELD("Triggers-Pending"), f_trigpend,        w_trigpend                                 },
+  { FIELD("Triggers-Awaited"), f_trigaw,          w_trigaw                                   },
 
-static const struct nickname nicknames[] = {
-  /* Note: Capitalization of these strings is important. */
-  { .nick = "Recommended",      .canon = "Recommends" },
-  { .nick = "Optional",         .canon = "Suggests" },
-  { .nick = "Class",            .canon = "Priority" },
-  { .nick = "Package-Revision", .canon = "Revision" },
-  { .nick = "Package_Revision", .canon = "Revision" },
-  { .nick = NULL }
+  /* The following are the obsolete fields that get remapped to their
+   * modern forms, while emitting an obsolescence warning. */
+  { FIELD("Recommended"),      f_obs_dependency,  w_null,           dep_recommends           },
+  { FIELD("Optional"),         f_obs_dependency,  w_null,           dep_suggests             },
+  { FIELD("Class"),            f_obs_class,       w_null                                     },
+  { FIELD("Revision"),         f_obs_revision,    w_null                                     },
+  { FIELD("Package-Revision"), f_obs_revision,    w_null                                     },
+  { FIELD("Package_Revision"), f_obs_revision,    w_null                                     },
+  { NULL                                                                                     }
 };
 
 /**
@@ -117,26 +115,17 @@ pkg_parse_field(struct parsedb_state *ps, struct field_state *fs,
                 void *parse_obj)
 {
   struct pkg_parse_object *pkg_obj = parse_obj;
-  const struct nickname *nick;
   const struct fieldinfo *fip;
   int *ip;
 
-  for (nick = nicknames; nick->nick; nick++)
-    if (strncasecmp(nick->nick, fs->fieldstart, fs->fieldlen) == 0 &&
-        nick->nick[fs->fieldlen] == '\0')
-      break;
-  if (nick->nick) {
-    fs->fieldstart = nick->canon;
-    fs->fieldlen = strlen(fs->fieldstart);
-  }
-
   for (fip = fieldinfos, ip = fs->fieldencountered; fip->name; fip++, ip++)
-    if (strncasecmp(fip->name, fs->fieldstart, fs->fieldlen) == 0)
+    if (fip->namelen == (size_t)fs->fieldlen &&
+        strncasecmp(fip->name, fs->fieldstart, fs->fieldlen) == 0)
       break;
   if (fip->name) {
     if ((*ip)++)
       parse_error(ps,
-                  _("duplicate value for `%s' field"), fip->name);
+                  _("duplicate value for '%s' field"), fip->name);
 
     varbuf_reset(&fs->value);
     varbuf_add_buf(&fs->value, fs->valuestart, fs->valuelen);
@@ -148,17 +137,18 @@ pkg_parse_field(struct parsedb_state *ps, struct field_state *fs,
 
     if (fs->fieldlen < 2)
       parse_error(ps,
-                  _("user-defined field name `%.*s' too short"),
+                  _("user-defined field name '%.*s' too short"),
                   fs->fieldlen, fs->fieldstart);
     larpp = &pkg_obj->pkgbin->arbs;
     while ((arp = *larpp) != NULL) {
-      if (strncasecmp(arp->name, fs->fieldstart, fs->fieldlen) == 0)
+      if (strncasecmp(arp->name, fs->fieldstart, fs->fieldlen) == 0 &&
+          strlen(arp->name) == (size_t)fs->fieldlen)
         parse_error(ps,
-                   _("duplicate value for user-defined field `%.*s'"),
+                   _("duplicate value for user-defined field '%.*s'"),
                    fs->fieldlen, fs->fieldstart);
       larpp = &arp->next;
     }
-    arp = nfmalloc(sizeof(struct arbitraryfield));
+    arp = nfmalloc(sizeof(*arp));
     arp->name = nfstrnsave(fs->fieldstart, fs->fieldlen);
     arp->value = nfstrnsave(fs->valuestart, fs->valuelen);
     arp->next = NULL;
@@ -176,7 +166,7 @@ pkg_parse_verify(struct parsedb_state *ps,
   struct dependency *dep;
   struct deppossi *dop;
 
-  parse_must_have_field(ps, pkg->set->name, "package name");
+  parse_must_have_field(ps, pkg->set->name, "Package");
 
   /* XXX: We need to check for status != PKG_STAT_HALFINSTALLED as while
    * unpacking an unselected package, it will not have yet all data in
@@ -185,9 +175,9 @@ pkg_parse_verify(struct parsedb_state *ps,
   if ((ps->flags & pdb_recordavailable) ||
       (pkg->status != PKG_STAT_NOTINSTALLED &&
        pkg->status != PKG_STAT_HALFINSTALLED)) {
-    parse_ensure_have_field(ps, &pkgbin->description, "description");
-    parse_ensure_have_field(ps, &pkgbin->maintainer, "maintainer");
-    parse_must_have_field(ps, pkgbin->version.version, "version");
+    parse_ensure_have_field(ps, &pkgbin->description, "Description");
+    parse_ensure_have_field(ps, &pkgbin->maintainer, "Maintainer");
+    parse_must_have_field(ps, pkgbin->version.version, "Version");
   }
 
   /* XXX: Versions before dpkg 1.10.19 did not preserve the Architecture
@@ -196,12 +186,12 @@ pkg_parse_verify(struct parsedb_state *ps,
   if ((ps->flags & pdb_recordavailable) ||
       pkg->status > PKG_STAT_HALFINSTALLED) {
     /* We always want usable architecture information (as long as the package
-     * is in such a state that it make sense), so that it can be used safely
+     * is in such a state that it makes sense), so that it can be used safely
      * on string comparisons and the like. */
     if (pkgbin->arch->type == DPKG_ARCH_NONE)
-      parse_warn(ps, _("missing %s"), "architecture");
+      parse_warn(ps, _("missing '%s' field"), "Architecture");
     else if (pkgbin->arch->type == DPKG_ARCH_EMPTY)
-      parse_warn(ps, _("empty value for %s"), "architecture");
+      parse_warn(ps, _("empty value for '%s' field"), "Architecture");
   }
   /* Mark missing architectures as empty, to distinguish these from
    * unused slots in the db. */
@@ -210,12 +200,15 @@ pkg_parse_verify(struct parsedb_state *ps,
 
   if (pkgbin->arch->type == DPKG_ARCH_EMPTY &&
       pkgbin->multiarch == PKG_MULTIARCH_SAME)
-    parse_error(ps, _("package has field '%s' but is missing architecture"),
+    parse_error(ps, _("package has '%s' field but is missing architecture"),
                 "Multi-Arch: same");
   if (pkgbin->arch->type == DPKG_ARCH_ALL &&
       pkgbin->multiarch == PKG_MULTIARCH_SAME)
-    parse_error(ps, _("package has field '%s' but is architecture all"),
-                "Multi-Arch: same");
+    parse_error(ps, _("package has '%s' field but is architecture '%s'"),
+                "Multi-Arch: same", "all");
+
+  /* Generate the cached fully qualified package name representation. */
+  pkgbin->pkgname_archqual = pkgbin_name_archqual(pkg, pkgbin);
 
   /* Initialize deps to be arch-specific unless stated otherwise. */
   for (dep = pkgbin->depends; dep; dep = dep->next)
@@ -223,19 +216,25 @@ pkg_parse_verify(struct parsedb_state *ps,
       if (!dop->arch)
         dop->arch = pkgbin->arch;
 
-  /* Check the Config-Version information:
-   * If there is a Config-Version it is definitely to be used, but
-   * there shouldn't be one if the package is ‘installed’ (in which case
-   * the Version and/or Revision will be copied) or if the package is
-   * ‘not-installed’ (in which case there is no Config-Version). */
+  /*
+   * Check the Config-Version information:
+   *
+   * If there is a Config-Version it is definitely to be used, but there
+   * should not be one if the package is ‘installed’ or ‘triggers-pending’
+   * (in which case the Version will be copied) or if the package is
+   * ‘not-installed’ (in which case there is no Config-Version).
+   */
   if (!(ps->flags & pdb_recordavailable)) {
     if (pkg->configversion.version) {
       if (pkg->status == PKG_STAT_INSTALLED ||
-          pkg->status == PKG_STAT_NOTINSTALLED)
+          pkg->status == PKG_STAT_NOTINSTALLED ||
+          pkg->status == PKG_STAT_TRIGGERSPENDING)
         parse_error(ps,
-                    _("Config-Version for package with inappropriate Status"));
+                    _("'%s' field present for package with inappropriate '%s' field"),
+                    "Config-Version", "Status");
     } else {
-      if (pkg->status == PKG_STAT_INSTALLED)
+      if (pkg->status == PKG_STAT_INSTALLED ||
+          pkg->status == PKG_STAT_TRIGGERSPENDING)
         pkg->configversion = pkgbin->version;
     }
   }
@@ -248,7 +247,8 @@ pkg_parse_verify(struct parsedb_state *ps,
                 pkg_status_name(pkg));
   else if (pkg->status == PKG_STAT_TRIGGERSAWAITED && !pkg->trigaw.head)
     parse_error(ps,
-                _("package has status triggers-awaited but no triggers awaited"));
+                _("package has status %s but no triggers awaited"),
+                pkg_status_name(pkg));
 
   if (pkg->trigpend_head &&
       !(pkg->status == PKG_STAT_TRIGGERSPENDING ||
@@ -258,35 +258,32 @@ pkg_parse_verify(struct parsedb_state *ps,
                 pkg_status_name(pkg));
   else if (pkg->status == PKG_STAT_TRIGGERSPENDING && !pkg->trigpend_head)
     parse_error(ps,
-                _("package has status triggers-pending but no triggers "
-                  "pending"));
+                _("package has status %s but no triggers pending"),
+                pkg_status_name(pkg));
 
-  /* FIXME: There was a bug that could make a not-installed package have
+  /* Note: There was a bug that could make a not-installed package have
    * conffiles, so we check for them here and remove them (rather than
    * calling it an error, which will do at some point). */
   if (!(ps->flags & pdb_recordavailable) &&
       pkg->status == PKG_STAT_NOTINSTALLED &&
       pkgbin->conffiles) {
     parse_warn(ps,
-               _("Package which in state not-installed has conffiles, "
-                 "forgetting them"));
+               _("package has status %s and has conffiles, forgetting them"),
+               pkg_status_name(pkg));
     pkgbin->conffiles = NULL;
   }
 
-  /* XXX: Mark not-installed leftover packages for automatic removal on
-   * next database dump. This code can be removed after dpkg 1.16.x, when
-   * there's guarantee that no leftover is found on the status file on
-   * major distributions. */
+  /* Note: Mark not-installed leftover packages for automatic removal on
+   * next database dump. */
   if (!(ps->flags & pdb_recordavailable) &&
       pkg->status == PKG_STAT_NOTINSTALLED &&
       pkg->eflag == PKG_EFLAG_OK &&
       (pkg->want == PKG_WANT_PURGE ||
-       pkg->want == PKG_WANT_DEINSTALL ||
-       pkg->want == PKG_WANT_HOLD)) {
+       pkg->want == PKG_WANT_DEINSTALL)) {
     pkg_set_want(pkg, PKG_WANT_UNKNOWN);
   }
 
-  /* XXX: Mark not-installed non-arch-qualified selections for automatic
+  /* Note: Mark not-installed non-arch-qualified selections for automatic
    * removal, as they do not make sense in a multiarch enabled world, and
    * might cause those selections to be unreferencable from command-line
    * interfaces when there's other more specific selections. */
@@ -295,7 +292,18 @@ pkg_parse_verify(struct parsedb_state *ps,
       pkg->eflag == PKG_EFLAG_OK &&
       pkg->want == PKG_WANT_INSTALL &&
       pkgbin->arch->type == DPKG_ARCH_EMPTY)
-    pkg->want = PKG_WANT_UNKNOWN;
+    pkg_set_want(pkg, PKG_WANT_UNKNOWN);
+
+  /* Note: Versions before dpkg 1.13.10 did not blank the Origin and Bugs
+   * fields, so there can be packages that should be garbage collected but
+   * are lingering around. Blank them to make sure we will forget all about
+   * them on the next database dump. */
+  if (!(ps->flags & pdb_recordavailable) &&
+      pkg->status == PKG_STAT_NOTINSTALLED &&
+      pkg->eflag == PKG_EFLAG_OK &&
+      pkg->want == PKG_WANT_UNKNOWN) {
+    pkgbin_blank(pkgbin);
+  }
 }
 
 struct pkgcount {
@@ -334,7 +342,7 @@ parse_find_set_slot(struct parsedb_state *ps,
   struct pkgset *set;
   struct pkginfo *pkg;
 
-  set = pkg_db_find_set(new_pkg->set->name);
+  set = pkg_hash_find_set(new_pkg->set->name);
 
   /* Sanity checks: verify that the db is in a consistent state. */
 
@@ -397,9 +405,9 @@ parse_find_pkg_slot(struct parsedb_state *ps,
      * possible architecture switch, for example from native to all. */
     if (pkgset_installed_instances(db_set) == 1 &&
         new_pkgbin->multiarch != PKG_MULTIARCH_SAME)
-      return pkg_db_get_singleton(db_set);
+      return pkg_hash_get_singleton(db_set);
     else
-      return pkg_db_get_pkg(db_set, new_pkgbin->arch);
+      return pkg_hash_get_pkg(db_set, new_pkgbin->arch);
   } else {
     bool selection = false;
 
@@ -411,28 +419,29 @@ parse_find_pkg_slot(struct parsedb_state *ps,
     /* Verify we don't allow something that will mess up the db. */
     if (pkgset_installed_instances(db_set) > 1 &&
         !selection && new_pkgbin->multiarch != PKG_MULTIARCH_SAME)
-      ohshit(_("%s %s (Multi-Arch: %s) is not co-installable with "
-               "%s which has multiple installed instances"),
+      ohshit(_("package %s (%s) with field '%s: %s' is not co-installable "
+               "with %s which has multiple installed instances"),
              pkgbin_name(new_pkg, new_pkgbin, pnaw_always),
              versiondescribe(&new_pkgbin->version, vdew_nonambig),
-             multiarchinfos[new_pkgbin->multiarch].name, db_set->name);
+             "Multi-Arch", multiarchinfos[new_pkgbin->multiarch].name,
+             db_set->name);
 
     /* If we are parsing the status file, use a slot per arch. */
     if (ps->type == pdb_file_status)
-      return pkg_db_get_pkg(db_set, new_pkgbin->arch);
+      return pkg_hash_get_pkg(db_set, new_pkgbin->arch);
 
     /* If we are doing an update, from the log or a new package, then
      * handle cross-grades. */
     if (pkgset_installed_instances(db_set) == 1) {
-      db_pkg = pkg_db_get_singleton(db_set);
+      db_pkg = pkg_hash_get_singleton(db_set);
 
       if (db_pkg->installed.multiarch == PKG_MULTIARCH_SAME &&
           new_pkgbin->multiarch == PKG_MULTIARCH_SAME)
-        return pkg_db_get_pkg(db_set, new_pkgbin->arch);
+        return pkg_hash_get_pkg(db_set, new_pkgbin->arch);
       else
         return db_pkg;
     } else {
-      return pkg_db_get_pkg(db_set, new_pkgbin->arch);
+      return pkg_hash_get_pkg(db_set, new_pkgbin->arch);
     }
   }
 }
@@ -472,18 +481,21 @@ pkg_parse_copy(struct parsedb_state *ps,
     pkg_copy_eflags(dst_pkg, src_pkg);
     pkg_set_status(dst_pkg, src_pkg->status);
     dst_pkg->configversion = src_pkg->configversion;
-    dst_pkg->files = NULL;
+    dst_pkg->archives = NULL;
 
     dst_pkg->trigpend_head = src_pkg->trigpend_head;
     dst_pkg->trigaw = src_pkg->trigaw;
     for (ta = dst_pkg->trigaw.head; ta; ta = ta->sameaw.next) {
-      assert(ta->aw == src_pkg);
+      if (ta->aw != src_pkg)
+        internerr("trigger awaited package %s and origin package %s not linked properly",
+                  pkg_name(ta->aw, pnaw_always),
+                  pkgbin_name(src_pkg, src_pkgbin, pnaw_always));
       ta->aw = dst_pkg;
-      /* ->othertrigaw_head is updated by trig_note_aw in *(pkg_db_find())
+      /* ->othertrigaw_head is updated by trig_note_aw in *(pkg_hash_find())
        * rather than in dst_pkg. */
     }
-  } else if (!(ps->flags & pdb_ignorefiles)) {
-    dst_pkg->files = src_pkg->files;
+  } else if (!(ps->flags & pdb_ignore_archives)) {
+    dst_pkg->archives = src_pkg->archives;
   }
 }
 
@@ -515,11 +527,16 @@ parsedb_new(const char *filename, int fd, enum parsedbflags flags)
   struct parsedb_state *ps;
 
   ps = m_malloc(sizeof(*ps));
+  ps->err = DPKG_ERROR_OBJECT;
+  ps->errmsg = VARBUF_OBJECT;
   ps->filename = filename;
   ps->type = parse_get_type(ps, flags);
   ps->flags = flags;
   ps->fd = fd;
   ps->lno = 0;
+  ps->data = NULL;
+  ps->dataptr = NULL;
+  ps->endptr = NULL;
   ps->pkg = NULL;
   ps->pkgbin = NULL;
 
@@ -540,13 +557,13 @@ parsedb_open(const char *filename, enum parsedbflags flags)
     return parsedb_new(filename, STDIN_FILENO, flags);
 
   fd = open(filename, O_RDONLY);
-  if (fd == -1)
-    ohshite(_("failed to open package info file `%.255s' for reading"),
+  if (fd == -1 && !(errno == ENOENT && (flags & pdb_allow_empty)))
+    ohshite(_("failed to open package info file '%.255s' for reading"),
             filename);
 
   ps = parsedb_new(filename, fd, flags | pdb_close_fd);
 
-  push_cleanup(cu_closefd, ~ehflag_normaltidy, NULL, 0, 1, &ps->fd);
+  push_cleanup(cu_closefd, ~ehflag_normaltidy, 1, &ps->fd);
 
   return ps;
 }
@@ -559,8 +576,11 @@ parsedb_load(struct parsedb_state *ps)
 {
   struct stat st;
 
+  if (ps->fd < 0 && (ps->flags & pdb_allow_empty))
+      return;
+
   if (fstat(ps->fd, &st) == -1)
-    ohshite(_("can't stat package info file `%.255s'"), ps->filename);
+    ohshite(_("can't stat package info file '%.255s'"), ps->filename);
 
   if (S_ISFIFO(st.st_mode)) {
     struct varbuf buf = VARBUF_INIT;
@@ -579,7 +599,7 @@ parsedb_load(struct parsedb_state *ps)
 #ifdef USE_MMAP
     ps->dataptr = mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, ps->fd, 0);
     if (ps->dataptr == MAP_FAILED)
-      ohshite(_("can't mmap package info file `%.255s'"), ps->filename);
+      ohshite(_("can't mmap package info file '%.255s'"), ps->filename);
 #else
     ps->dataptr = m_malloc(st.st_size);
 
@@ -603,7 +623,7 @@ parse_stanza(struct parsedb_state *ps, struct field_state *fs,
   int c;
 
   /* Skip adjacent new lines. */
-  while (!parse_EOF(ps)) {
+  while (!parse_at_eof(ps)) {
     c = parse_getc(ps);
     if (c != '\n' && c != MSDOS_EOF_CHAR)
       break;
@@ -611,7 +631,7 @@ parse_stanza(struct parsedb_state *ps, struct field_state *fs,
   }
 
   /* Nothing relevant parsed, bail out. */
-  if (parse_EOF(ps))
+  if (parse_at_eof(ps))
     return false;
 
   /* Loop per field. */
@@ -620,7 +640,7 @@ parse_stanza(struct parsedb_state *ps, struct field_state *fs,
 
     /* Scan field name. */
     fs->fieldstart = ps->dataptr - 1;
-    while (!parse_EOF(ps) && !isspace(c) && c != ':' && c != MSDOS_EOF_CHAR)
+    while (!parse_at_eof(ps) && !c_isspace(c) && c != ':' && c != MSDOS_EOF_CHAR)
       c = parse_getc(ps);
     fs->fieldlen = ps->dataptr - fs->fieldstart - 1;
     if (fs->fieldlen == 0)
@@ -630,38 +650,35 @@ parse_stanza(struct parsedb_state *ps, struct field_state *fs,
                   fs->fieldlen, fs->fieldstart);
 
     /* Skip spaces before ‘:’. */
-    while (!parse_EOF(ps) && c != '\n' && isspace(c))
+    while (!parse_at_eof(ps) && c != '\n' && c_isspace(c))
       c = parse_getc(ps);
 
     /* Validate ‘:’. */
-    if (parse_EOF(ps))
-      parse_error(ps,
-                  _("EOF after field name `%.*s'"), fs->fieldlen, fs->fieldstart);
+    if (parse_at_eof(ps))
+      parse_error(ps, _("end of file after field name '%.*s'"),
+                  fs->fieldlen, fs->fieldstart);
     if (c == '\n')
       parse_error(ps,
-                  _("newline in field name `%.*s'"), fs->fieldlen, fs->fieldstart);
+                  _("newline in field name '%.*s'"), fs->fieldlen, fs->fieldstart);
     if (c == MSDOS_EOF_CHAR)
-      parse_error(ps,
-                  _("MSDOS EOF (^Z) in field name `%.*s'"),
+      parse_error(ps, _("MSDOS end of file (^Z) in field name '%.*s'"),
                   fs->fieldlen, fs->fieldstart);
     if (c != ':')
       parse_error(ps,
-                  _("field name `%.*s' must be followed by colon"),
+                  _("field name '%.*s' must be followed by colon"),
                   fs->fieldlen, fs->fieldstart);
 
     /* Skip space after ‘:’ but before value and EOL. */
-    while (!parse_EOF(ps)) {
+    while (!parse_at_eof(ps)) {
       c = parse_getc(ps);
-      if (c == '\n' || !isspace(c))
+      if (c == '\n' || !c_isspace(c))
         break;
     }
-    if (parse_EOF(ps))
-      parse_error(ps,
-                  _("EOF before value of field `%.*s' (missing final newline)"),
+    if (parse_at_eof(ps))
+      parse_error(ps, _("end of file before value of field '%.*s' (missing final newline)"),
                   fs->fieldlen, fs->fieldstart);
     if (c == MSDOS_EOF_CHAR)
-      parse_error(ps,
-                  _("MSDOS EOF char in value of field `%.*s' (missing newline?)"),
+      parse_error(ps, _("MSDOS end of file (^Z) in value of field '%.*s' (missing newline?)"),
                   fs->fieldlen, fs->fieldstart);
 
     blank_line = false;
@@ -670,29 +687,32 @@ parse_stanza(struct parsedb_state *ps, struct field_state *fs,
     fs->valuestart = ps->dataptr - 1;
     for (;;) {
       if (c == '\n' || c == MSDOS_EOF_CHAR) {
-        if (blank_line)
-          parse_error(ps,
-                      _("blank line in value of field '%.*s'"),
-                      fs->fieldlen, fs->fieldstart);
+        if (blank_line) {
+          if (ps->flags & pdb_lax_stanza_parser)
+            parse_warn(ps, _("blank line in value of field '%.*s'"),
+                       fs->fieldlen, fs->fieldstart);
+          else
+            parse_error(ps, _("blank line in value of field '%.*s'"),
+                        fs->fieldlen, fs->fieldstart);
+        }
         ps->lno++;
 
-        if (parse_EOF(ps))
+        if (parse_at_eof(ps))
           break;
         c = parse_getc(ps);
 
         /* Found double EOL, or start of new field. */
-        if (parse_EOF(ps) || c == '\n' || !isspace(c))
+        if (parse_at_eof(ps) || c == '\n' || !c_isspace(c))
           break;
 
         parse_ungetc(c, ps);
         blank_line = true;
-      } else if (blank_line && !isspace(c)) {
+      } else if (blank_line && !c_isspace(c)) {
         blank_line = false;
       }
 
-      if (parse_EOF(ps))
-        parse_error(ps,
-                    _("EOF during value of field `%.*s' (missing final newline)"),
+      if (parse_at_eof(ps))
+        parse_error(ps, _("end of file during value of field '%.*s' (missing final newline)"),
                     fs->fieldlen, fs->fieldstart);
 
       c = parse_getc(ps);
@@ -700,12 +720,12 @@ parse_stanza(struct parsedb_state *ps, struct field_state *fs,
     fs->valuelen = ps->dataptr - fs->valuestart - 1;
 
     /* Trim ending space on value. */
-    while (fs->valuelen && isspace(*(fs->valuestart + fs->valuelen - 1)))
+    while (fs->valuelen && c_isspace(*(fs->valuestart + fs->valuelen - 1)))
       fs->valuelen--;
 
     parse_field(ps, fs, parse_obj);
 
-    if (parse_EOF(ps) || c == '\n' || c == MSDOS_EOF_CHAR)
+    if (parse_at_eof(ps) || c == '\n' || c == MSDOS_EOF_CHAR)
       break;
   } /* Loop per field. */
 
@@ -724,8 +744,8 @@ parsedb_close(struct parsedb_state *ps)
   if (ps->flags & pdb_close_fd) {
     pop_cleanup(ehflag_normaltidy);
 
-    if (close(ps->fd))
-      ohshite(_("failed to close after read: `%.255s'"), ps->filename);
+    if (ps->fd >= 0 && close(ps->fd) < 0)
+      ohshite(_("failed to close after read: '%.255s'"), ps->filename);
   }
 
   if (ps->data != NULL) {
@@ -735,6 +755,8 @@ parsedb_close(struct parsedb_state *ps)
     free(ps->data);
 #endif
   }
+  dpkg_error_destroy(&ps->err);
+  varbuf_destroy(&ps->errmsg);
   free(ps);
 }
 
@@ -755,6 +777,9 @@ parsedb_parse(struct parsedb_state *ps, struct pkginfo **donep)
   int pdone;
   struct field_state fs;
 
+  if (ps->data == NULL && (ps->flags & pdb_allow_empty))
+    return 0;
+
   memset(&fs, 0, sizeof(fs));
   fs.fieldencountered = fieldencountered;
 
@@ -802,13 +827,13 @@ parsedb_parse(struct parsedb_state *ps, struct pkginfo **donep)
     if (donep)
       *donep = db_pkg;
     pdone++;
-    if (parse_EOF(ps))
+    if (parse_at_eof(ps))
       break;
   }
 
   varbuf_destroy(&fs.value);
   if (donep && !pdone)
-    ohshit(_("no package information in `%.255s'"), ps->filename);
+    ohshit(_("no package information in '%.255s'"), ps->filename);
 
   return pdone;
 }
@@ -853,7 +878,7 @@ parsedb(const char *filename, enum parsedbflags flags, struct pkginfo **pkgp)
  * It is likely that the backward pointer for the package in question
  * (‘depended’) will be updated by this routine, but this will happen by
  * the routine traversing the dependency data structures. It doesn't need
- * to be told where to update that; I just mention it as something that
+ * to be told where to update that; just mentioned here as something that
  * one should be cautious about.
  */
 void copy_dependency_links(struct pkginfo *pkg,
